/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
/*
 * Ca.java,v 1.2 2002/08/02 15:39:10 jba Exp
 */

package jca;

import jca.event.*;
import jca.dbr.*;
import java.lang.*;
import java.util.*;

// For jdk 1.1.7 compatibility
//import javax.swing.*;



/**
 * The Java Channel Access class.
 * 
 * @author Eric BOUCHER
 * @see jca.PV
 * @see jca.Monitor
 */
public final class Ca {

  /*
   * Channel Access OP code. Get request.
   */
  public static final int OP_GET        = 0;
  /*
   * Channel Access OP code. Put request.
   */
  public static final int OP_PUT        = 1;
  /*
   * Channel Access OP code. Search request.
   */
  public static final int OP_SEARCH     = 2;
  /*
   * Channel Access OP code. Add event request.
   */
  public static final int OP_ADD_EVENT  = 3;
  /*
   * Channel Access OP code. Clear event request.
   */
  public static final int OP_CLEAR_EVENT= 4;
  /*
   * Channel Access OP code. Other request.
   */
  public static final int OP_OTHER      = 5;
  /*
   * Channel Access OP code. Connection up notification.
   */
  public static final int OP_CONN_UP    = 6;
  /*
   * Channel Access OP code. Connection down notification.
   */
  public static final int OP_CONN_DOWN  = 7;


  /*
   * Load the Channel Access library.
   */
  static {
    System.loadLibrary("jca");
  }

  /**
   * Initializes the Channel Access threads.
   */
  public static synchronized void init() {
    if(m_event==null) m_event= new EventProcessor();
    if(m_request==null) m_request= new RequestProcessor();
  }

  /**
   * Frees all CA resources and exits the Channel Access thread
   */
  public static void exit() {
    try {
      if(m_event!=null) {
        m_event.exit();
        m_event.join();
        m_event= null;
      }
      if(m_request!=null) {
        //       m_request.process(new ExitRequest());
        m_request.exit();
        m_request.join();
        m_request= null;
      }
    } catch(InterruptedException e) {
    }
  }

  /**
   * Sends all buffered IO requests. Usually performed by pendIO or pendEvent.
   * However, some users may wish to flush remote operation requests early so
   * that they can perform operations in parallel with outstanding remote
   * operations.
   *
   * @see       jca.Ca#pendEvent
   * @see       jca.Ca#pendIO
   */
  public static void flushIO() {
    if(m_request==null) return;
    m_request.process(new FlushIORequest());
  }

  /**
   * Flushes the send buffer and waits until outstanding queries complete.
   * At this time queries include calls to PV(name) and PV.get(value). 
   * <p>
   * Comments: Any values written into your program's variables by a CA synchronous
   * request should not be referenced by your program until Ca.pendIO() has been called.
   *
   * @param  timeout    time in seconds to wait before the function times out. A timeout 
   *                    of zero means wait forever.
   * @exception jca.TimeOutException    The timeout has expire and all outstanding queries
   *                                    were not completed.
   */
  public static void pendIO(double timeout) throws jca.TimeOutException {
    if(m_request==null) return;
    if(m_request.process(new PendIORequest(timeout))!=ECA_NORMAL) 
      throw new jca.TimeOutException();
  }

  /**
   * Flushes the send buffer and waits for asynchronous event. Asynchronous event
   * include all call to PV(name,listener), PV.get(value,listener), PV.put(value,listener)
   * and all monitors events.
   * This routine will not return before the period expires and all unfinished 
   * channel access labor has been processed.
   * <p>
   * Comments: A very short timeout such as 0.0001 seconds will result in a poll
   *           (i.e. channel access will just check for file descriptor activity
   *           and then return).
   *
   * @param  period     time in seconds to wait before the function returns. A period 
   *                    of zero means wait forever.
   */
  public static void pendEvent(double period) {
    if(m_request==null) return;
    m_request.process(new PendEventRequest(period));
  }



  protected static void searchPV(jca.PV pv, jca.event.ConnectionListener l) {
    if(pv.id()<0) {
      init();
      jca.event.ConnectionEvent event= null;
      if(l!=null) {
        event= new jca.event.ConnectionEvent(pv,l);
      }
      m_request.process(new SearchRequest(pv, event));
    }
  }

  protected static void clearPV(jca.PV pv) {
    if(m_request==null) return;
    m_request.process(new ClearRequest(pv));
  }

  protected static void put(jca.PV pv, jca.dbr.DBR value) 
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    if(m_request==null) return;
    switch(m_request.process(new PutRequest(pv,value))) {
    case ECA_BADTYPE: throw new jca.BadTypeException();
    case ECA_BADCOUNT: throw new jca.BadCountException();
    case ECA_PUTFAIL: throw new jca.PutFailException();
    }
  }

  protected static void put(jca.PV pv, jca.dbr.DBR value, jca.event.PutListener listener,
                            Object arg)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    if(m_request==null) return;
    jca.event.PutEvent event= new jca.event.PutEvent(pv,value,arg,listener);
    switch(m_request.process(new PutCallbackRequest(event))) {
    case ECA_BADTYPE: throw new jca.BadTypeException();
    case ECA_BADCOUNT: throw new jca.BadCountException();
    case ECA_PUTFAIL: throw new jca.PutFailException();
    }
  }

  protected static void get(jca.PV pv, jca.dbr.DBR value)
    throws jca.BadTypeException, jca.BadCountException, jca.GetFailException {
    if(m_request==null) return;
    switch(m_request.process(new GetRequest(pv, value))) {
    case ECA_BADTYPE: throw new jca.BadTypeException();
    case ECA_BADCOUNT: throw new jca.BadCountException();
    case ECA_GETFAIL: throw new jca.GetFailException();
    }
  }

  protected static void get(jca.PV pv, jca.dbr.DBR value, jca.event.GetListener listener,
                            Object arg)
    throws jca.BadTypeException, jca.BadCountException, jca.GetFailException {
    if(m_request==null) return;
    jca.event.GetEvent event= new jca.event.GetEvent(pv,value,arg,listener);
    switch(m_request.process(new GetCallbackRequest(event))) {
    case ECA_BADTYPE:  throw new jca.BadTypeException();
    case ECA_BADCOUNT: throw new jca.BadCountException();
    case ECA_GETFAIL:  throw new jca.GetFailException();
    }
  }

  protected static jca.Monitor addMonitor(jca.PV pv, jca.dbr.DBR value, 
                                          jca.event.MonitorListener listener,
                                          Object arg, int mask)
    throws jca.BadTypeException, jca.BadCountException, jca.MonitorFailException  {

    if(m_request==null) return null;
    jca.event.MonitorEvent event= new jca.event.MonitorEvent(pv,value,arg,listener);

    switch(m_request.process(new AddMonitorRequest(event,mask))) {
    case ECA_BADTYPE: throw new jca.BadTypeException();
    case ECA_BADCOUNT: throw new jca.BadCountException();
    case ECA_ADDFAIL: throw new jca.MonitorFailException();
    }
    return new jca.Monitor(event);
  }

  protected static void clearMonitor(jca.Monitor monitor) {
    if(m_request==null) return;
    m_request.process(new ClearMonitorRequest(monitor));
  }

  protected static void setConnectionListener(jca.PV pv, jca.event.ConnectionListener l) {
    if(m_request==null) return;
    jca.event.ConnectionEvent event= new jca.event.ConnectionEvent(pv,l);
    m_request.process(new ConnectListenerRequest(event));
  }

  public static void queueEvent(CaEvent e, int status) {
    if(m_event!=null) {
      e.update(status);
      m_event.queueEvent(e);
    }
  }

  public static int queueSize() {
    if(m_event!=null) {
      return m_event.queueSize();
    }
    return 0;
  }

  public static long eventProcessed() {
    return m_event.eventProcessed();
  }

  protected void finalize() throws Throwable { 
    exit();
    super.finalize();
  }

  protected native static final 
    int jni_init();

  protected native static final 
    int jni_exit();

  protected native static final 
    int jni_search(jca.PV pv, jca.event.ConnectionEvent event);

  protected native static final
    int jni_clear_channel(int jpv);

  protected native static final
    int jni_flush_io();

  protected native static final
    int jni_pend_event(double timeout);

  protected native static final
    int jni_pend_io(double timeout);

  protected native static final
    int jni_put(int type, int count, int id, jca.dbr.DBR value);

  protected native static final
    int jni_put_callback(int type, int count, int id,
                         jca.dbr.DBR dbr, jca.event.PutEvent event);

  protected native static final
    int jni_get(int type, int count, int id, jca.dbr.DBR dbr);

  protected native static final
    int jni_get_callback(int type, int count, int id,
                         jca.dbr.DBR dbr, jca.event.GetEvent event);

  protected native static final
    int jni_add_monitor(int type, int count, int id,
                        jca.dbr.DBR dbr, jca.event.MonitorEvent event, int mask);

  protected native static final
    int jni_clear_monitor(int id);

  protected native static final 
    int jni_connection_listener(int id, jca.event.ConnectionEvent event);


  private static RequestProcessor m_request= null; 
  private static EventProcessor   m_event= null; 


  private static final int ECA_NORMAL    = 1;
  private static final int ECA_BADTYPE   = 114;
  private static final int ECA_BADCOUNT  = 176;
  private static final int ECA_GETFAIL   = 152;
  private static final int ECA_PUTFAIL   = 160;
  private static final int ECA_ADDFAIL   = 168;
}


/*
 * Thread in charge of dispatching Channel Access events.
 *
 * @see jca.event.CaEvent
 */
class EventProcessor extends Thread {
// for jdk1.1.7 compatibility
//   private LinkedList   m_queue;
  private Vector       m_queue;
  private long         m_count;
  private boolean      m_exit;
  

  public EventProcessor() {
    m_exit= false;
    m_count= 0;
//     m_queue= new LinkedList();
    m_queue= new Vector();
    setPriority(Thread.NORM_PRIORITY+1);
    start();
  }

  public void exit() {
    m_exit= true;
  }

  public void queueEvent(CaEvent e) {
    synchronized(m_queue) {
//       m_queue.addLast(e);
      m_queue.addElement(e);
      m_queue.notifyAll();
    }
  }

  public int queueSize() {
    return m_queue.size();
  }

  public long eventProcessed() {
    return m_count;
  }

  public void run() {
    CaEvent event;
    try {
      while(!m_exit) {
        synchronized(m_queue) {
          while(m_queue.size()==0 && !m_exit) {
            m_queue.wait(100);
          }
        }
        if(!m_exit) {
          synchronized(m_queue) {
//             event=(CaEvent)(m_queue.removeFirst());
            event=(CaEvent)(m_queue.firstElement());
            m_queue.removeElement(event);
          }
          event.fire();
          ++m_count;
          yield();
        }
      }
    } catch(InterruptedException e) {
    } finally {
    }
  }
}

/*
 * Thread in charge of sending Channel Access request to the JNI.
 *
 */
class RequestProcessor extends Thread {

  private Request m_request;
  private boolean m_exit;

  public RequestProcessor() {
    m_exit= false;
    m_request= null;
    setPriority(Thread.NORM_PRIORITY+2);
    start();
  }

  public void exit() {
    m_exit= true;
  }

  public synchronized int process(Request request) {
    try {
      while(m_request!=null) {
        wait(100);
      }
      m_request= request;
      notifyAll();
      while(m_request!=null) {
        wait(100);
      }
      return request.status;
    } catch(InterruptedException e) {
      return 0;
    }
  }

  public synchronized void run() {
    Ca.jni_init();
    try {
      while(m_exit!=true) {
        while(m_request==null && m_exit!=true) {
          // Wait for something to do
          wait(100);
        }
        if(m_exit!=true) {
          m_request.process();
          // Notify that the request has been processed
          m_request= null;
          notifyAll();
        }
      }
    } catch(InterruptedException e) {
    } finally {
      Ca.jni_exit();
    }
  }  
}



abstract class Request {
  public int status;
  abstract public void process();
}

class SearchRequest extends jca.Request {
  private jca.PV m_pv;
  private jca.event.ConnectionEvent m_event;

  public SearchRequest(jca.PV pv, jca.event.ConnectionEvent event) {
    m_pv= pv;
    m_event= event;
  }

  public void process() {
    status= Ca.jni_search(m_pv,m_event);
  }
}


class ClearRequest extends jca.Request {
  private jca.PV m_pv;

  public ClearRequest(jca.PV pv) {
    m_pv= pv;
  }
  public void process() {
    status= Ca.jni_clear_channel(m_pv.id());
  }
}

class FlushIORequest extends jca.Request {
  public FlushIORequest() {
  }

  public void process() {
    status= Ca.jni_flush_io();
  }
}

class PendEventRequest extends jca.Request {
  private double m_timeout;

  public PendEventRequest(double timeout) {
    m_timeout= timeout;
  }
  public void process() {
    status= Ca.jni_pend_event(m_timeout);
  }
}

class PendIORequest extends jca.Request {
  private double m_timeout;

  public PendIORequest(double timeout) {
    m_timeout= timeout;
  }
  public void process() {
    status= Ca.jni_pend_io(m_timeout);
  }
}

class PutRequest extends jca.Request {
  private jca.PV      m_pv;
  private jca.dbr.DBR m_value;
  
  public PutRequest(jca.PV pv, jca.dbr.DBR value) {
    m_pv= pv;
    m_value= value;
  }
  public void process() {
    status= Ca.jni_put(m_value.type(), m_value.count(), m_pv.id(), m_value);
  }
}

class PutCallbackRequest extends jca.Request {
  private jca.event.PutEvent    m_event;

  public PutCallbackRequest(jca.event.PutEvent event) {
    m_event= event;
  }
  public void process() {
    status= Ca.jni_put_callback(m_event.type(), m_event.count(), m_event.pv().id(),
                                m_event.dbr(), m_event);
  }
}

class GetRequest extends jca.Request {
  private jca.PV      m_pv;
  private jca.dbr.DBR m_buffer;

  public GetRequest(jca.PV pv, jca.dbr.DBR buffer) {
    m_pv= pv;
    m_buffer= buffer;
  }
  public void process() {
    status= Ca.jni_get(m_buffer.type(), m_buffer.count(), m_pv.id(), m_buffer);
  }
}


class GetCallbackRequest extends jca.Request {
  private jca.event.GetEvent m_event;  

  public GetCallbackRequest(jca.event.GetEvent event) {
    m_event= event;
  }

  public void process() {
    status= Ca.jni_get_callback(m_event.type(), m_event.count(), m_event.pv().id(),
                                m_event.dbr(), m_event);
  }
}

class AddMonitorRequest extends jca.Request {
  private jca.event.MonitorEvent m_event;
  private int                    m_mask;

  public AddMonitorRequest(jca.event.MonitorEvent event, int mask) {
    m_event= event;
    m_mask= mask;
  }
  public void process() {
    status= Ca.jni_add_monitor(m_event.type(), m_event.count(), m_event.pv().id(),
                               m_event.dbr(), m_event, m_mask);
  }
}

class ClearMonitorRequest extends jca.Request {
  private jca.Monitor m_monitor;

  ClearMonitorRequest(jca.Monitor monitor) {
    m_monitor= monitor;
  }
  public void process() {
    status= Ca.jni_clear_monitor(m_monitor.id());
  }
}


class ConnectListenerRequest extends jca.Request {
  private jca.event.ConnectionEvent m_event;

  ConnectListenerRequest(jca.event.ConnectionEvent event) {
    m_event= event;
  }

  public void process() {
    status= Ca.jni_connection_listener(m_event.pv().id(), m_event);
  }
}




